/*\
|*| ---------------------------------------------------------------------------
|*|	Cirrus (integrated) ramdac driver
|*| ---------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#define	MAINTAINER	"Martin Erhardsen, [u930534@daimi.aau.dk]"
#define	DAC_VERSION	"0.05 ALPHA"

#ifdef DEBUG
#define	NEED_GRAPHTYPE_SYMS
#endif

#include <asm/io.h>
#include <ggi/graphdev.h>

#if !((MONITORVERSION == 01) && (MODEVERSION == 03) && (RAMDACVERSION == 01))
  #warning kgi datastructures have changed, check this code !!
#endif

#define	DAC_PRECISION	6	/* we can do 6 bit per RGB value	*/
#define	SHIFT (GGI_COLOR_PRECISION - DAC_PRECISION)
#define	COLOR(X) ((X << SHIFT) | (X >> (DAC_PRECISION - SHIFT)))

static void set_palette(ggi_palette *table)
{
  unsigned short *data,len;

  if (table && (data = (unsigned short *) table->data) && (len=3*table->cnt))
    { 
      outb(table->first,0x3C8);
      while (len--) 
	outb(*(data++) >> SHIFT,0x3C9);
    }
}

struct kgi_ramdac 
ramdac = {	"Cirrus",			/* manufacturer		*/
	 	"CL-GD542x",			/* model		*/
 		GM_ALL,         	        /* supported modes	*/
	 	0,			 	/* flags		*/
	 	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/

		 DAC_16BIT_PB,			/* device type		*/
		  {       0,  85226000 },	/* normal mode		*/
		  {       0,         0 },	/* 2xclk mode		*/

		 {{	  0,  32000000 },	/* 	GT_TEXT		*/
		  {       0,  85226000 },	/* 	GT_1BIT		*/
		  {	  0,         0 },	/* 	GT_4BIT		*/
		  {       0,  85226000 },	/* 	GT_8BIT		*/	
		  {	  0,  45076000 },	/* 	GT_15BIT	*/
		  { 	  0,  45076000 },	/* 	GT_16BIT	*/
		  {       0,  30000000 },	/* 	GT_24BIT	*/
		  {	  0,  45000000 },	/* 	GT_32BIT	*/
		  {	  0,  85226000 }}	/* 	GT_4BIT_PLANAR	*/
         };

/*\
|*|	GGI interface
\*/

int
kgi_ramdac_init(void)
{
  TRACE(printk("%s %s dac driver v"DAC_VERSION"\n", \
        ramdac.manufact, ramdac.model));
  return EOK;
}

void
kgi_ramdac_done(void)
{
  TRACE(printk("%s %s driver removed.\n", ramdac.manufact, ramdac.model));
}

struct kgi_ramdac
*kgi_ramdac_get_info (void)
{
  return &ramdac; 
}

int
kgi_ramdac_check_mode (struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int newclock = mode->dclk;

  if (!((1<<mode->request.graphtype)&(ramdac.modes)))
      {
        TRACE(printk("%s: could not handle graphtype %s\n",\
		     __FILE__, GRAPHMASK(mode->request.graphtype)));
        return -ERAMDAC;
      };
	
  
   switch (cmd)
   {
     case CMD_PROPOSE:
	mode->lclk.mul = mode->clk.mul = 1;
        mode->lclk.div = mode->clk.div = 1;
	return EOK;

     case CMD_LOWER:
	if (   (newclock < ramdac.clock[mode->request.graphtype].min)
	    || (newclock < ramdac.normal.min))
	  {
	    TRACE(printk("%s: DCLK too low (%i Hz).\n", __FILE__, newclock));
	    return -ERAMDAC;
	  }

	if (ramdac.clock[mode->request.graphtype].max < newclock)
    	  newclock = ramdac.clock[mode->request.graphtype].max;

	mode->dclk   = newclock;
	return EOK;

     case CMD_RAISE:
	if (   (newclock > ramdac.clock[mode->request.graphtype].max)
            || (newclock > ramdac.normal.max))
	  {
	    TRACE(printk("%s: DCLK too high (%i Hz).\n", __FILE__, newclock));
	    return -ERAMDAC;
	  }

	if (newclock < ramdac.clock[mode->request.graphtype].min)
	  newclock = ramdac.clock[mode->request.graphtype].min;

	if (newclock < ramdac.normal.min)
	  newclock = ramdac.normal.min;

	mode->dclk = newclock;
	return EOK;

     case CMD_CHECK:
	if (   (mode->lclk.mul != 1) || (mode->lclk.div != 1))
	  return -ERAMDAC;

	if (   (mode->dclk < ramdac.normal.min)
	    || (mode->dclk < ramdac.clock[mode->request.graphtype].min))
	  return -ERAMDAC;

	if (   (mode->dclk > ramdac.normal.max)
	    || (mode->dclk > ramdac.clock[mode->request.graphtype].max))
	  return -ERAMDAC;

	if (kgi_monitor_get_info()->synctype & COMPOSITE_SYNC)
	  {
	    TRACE(printk("%s: composite syncing not supported.\n",__FILE__));
	    return -ERAMDAC;
	  }

	return EOK;

    default:
	INTERNALERROR;
	return -ERAMDAC;
  }
}

void
kgi_ramdac_set_mode(struct kgi_mode *mode)
{
  outb(0xFF,0x3C6);
  set_palette(mode->palette);	/* set clut of mode */
}

int
kgi_ramdac_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
		  unsigned long arg)
{
  struct kgi_state *gr = file->private_data;
  
  switch(cmd)
    {
    case RAMDAC_GETINFO:
      _out_buf = &ramdac; _io_cnt = sizeof(ramdac);
      return EOK;
    case RAMDAC_SETINFO:
      if (suser())
	{
	  memcpy(&ramdac, _io_buf, sizeof(ramdac));
	  return EOK;
	}
      else
	return -EPERM;
   
    case RAMDAC_SETCLUT:
	{  ggi_palette *h = (void*) _io_buf;
	   int first,cnt;

	   first=h->first;cnt=h->cnt;
           if (first+cnt > 256) return -ENBOGUS;
           if (0==verify_area (VERIFY_READ, (void *) h->data, sizeof(ggi_color)*cnt))
           { memcpy_fromfs (&(gr->color[first]), 
           		    (void *) h->data, sizeof(ggi_color)*cnt); }
	   set_palette(gr->mode.palette);		/* set clut of mode */
	} 
	return 0;
    case RAMDAC_GETCLUT:
	{  ggi_palette *h = (void*) _io_buf;
	   int first,cnt;
	   _out_buf=_io_buf;

	   first=h->first;cnt=h->cnt;
           if (first+cnt > 256) return -ENBOGUS;
           if (0==verify_area (VERIFY_WRITE, (void *) h->data, sizeof(ggi_color)*cnt))
           { memcpy_tofs ((void *) h->data,
           		  &(gr->color[first]), sizeof(ggi_color)*cnt); }
	}
	return 0;
    default:
      return -ERAMDAC;
    }
}



